import json


def get_json_writer(type_: dict):
    def write_json(data: dict, path: str, **kwargs):
        json.dump(data, open(path, "w"), **kwargs)

    write_json.__sufix__ = ".json"
    return write_json


def get_json_reader(suffix=".json") -> dict:
    def read_json(path: str, **kwargs):
        return json.load(open(path, "r"), **kwargs)

    return read_json



from pathlib import Path

import click

from godata.project import create_project, list_collections, list_projects, load_project

from .utils import split_name


@click.command()
@click.argument("project_name")
@click.option(
    "--path",
    "-p",
    help="Path to the project's storage location. If not provided,"
    "the project will be created in the current directory.",
    default=lambda: Path.cwd(),
    type=Path,
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force creation of the project even if the path already exists.",
)
def create(project_name: str, path: Path, force: bool):
    """
    Create a project. The project's storage location will automatically be created.
    """
    name, collection = split_name(project_name)
    create_project(name, collection, path)


@click.command()
@click.argument("project_name")
@click.argument("project_path", type=str)
@click.argument("path", type=Path)
@click.option(
    "--recursive",
    "-r",
    is_flag=True,
    help="Only applies if linking a folder. If not set, this command will only "
    "link files in this particular folder and not any subfolders.",
)
@click.option(
    "--overwrite",
    "-o",
    is_flag=True,
    help="Force creation of the project even if something already exists.",
)
def link(
    project_name: str, project_path: str, path: Path, recursive: bool, overwrite: bool
):
    """
    Link a file or folder into a project.
    """
    name, collection = split_name(project_name)
    p = load_project(name, collection)
    p.link(path, project_path, recursive=recursive, overwrite=overwrite)


@click.command()
@click.argument("project_name", type=str)
@click.argument("project_path", type=str, required=False)
def ls(project_name: str, project_path: str = None):
    """
    List the contents of a project.
    """
    name, collection = split_name(project_name)
    p = load_project(name, collection)
    p.ls(project_path)


@click.command()
@click.argument("collection_name", type=str, required=False)
@click.option(
    "--hidden",
    "-h",
    is_flag=True,
    help="Include hidden projects or collections in the list.",
)
def list(collection_name: str = None, hidden: bool = False):
    """
    List the known collections, or the projects in a given collection
    """
    if collection_name is None:
        _ = list_collections(hidden, True)
    else:
        _ = list_projects(collection_name, hidden, True)


@click.command()
@click.argument("project_name", type=str)
@click.argument("project_path", type=str)
def get(project_name: str, project_path: str):
    """
    Get the real location of a file in a project.
    """
    name, collection = split_name(project_name)
    p = load_project(name, collection)
    path = p.get(project_path, as_path=True)
    click.echo(path)



__version__ = "0.8.1"
__minimum_server_version__ = "0.8.0"

from .ie import export_project, import_project
from .project import (
    create_project,
    delete_project,
    has_collection,
    has_project,
    list_collections,
    list_projects,
    load_project,
)

__all__ = [
    "load_project",
    "list_projects",
    "create_project",
    "list_collections",
    "delete_project",
    "has_project",
    "has_collection",
    "import_project",
    "export_project",
]



from pathlib import Path
from urllib import parse

import requests
from packaging import version

from godata import server

from .unixsocket import UnixHTTPAdapter

"""
The client connects to the godata server and communicates with it on behalf of the 
project. On Mac and Linux, this communication is done via a unix socket with a 
REST-ish API. Windows is not supported at this time, but will probalby just have 
to use a TCP socket instead of a unix socket.

The client is stateless, so it's just a bunch of functions.

I need to think a bit about how to properly reuse the client sesion.
"""
SERVER_PATH = str(Path.home() / ".godata.sock")
SERVER_URL = f"http+unix://{parse.quote(SERVER_PATH, safe='')}"
CLIENT = requests.Session()
CLIENT.mount("http+unix://", UnixHTTPAdapter(SERVER_PATH))


class NotFound(Exception):
    pass


class AlreadyExists(Exception):
    pass


class Forbidden(Exception):
    pass


class GodataClientError(Exception):
    pass


def check_server():
    from godata import __minimum_server_version__

    server_version = version.parse(get_version(CLIENT))
    minium_version = version.parse(__minimum_server_version__)

    if server_version < minium_version:
        raise GodataClientError(
            f"Server version {server_version} is less than minimum version "
            f"{minium_version}. Please upgrade the server."
        )
    return True


def get_client():
    if not hasattr(get_client, "has_run"):
        try:
            check_server()
        except requests.exceptions.ConnectionError:
            server.start()
            check_server()
        get_client.has_run = True
    return CLIENT


def get_version(client):
    resp = client.get(f"{SERVER_URL}/version")
    if resp.status_code == 200:
        return resp.json()


def list_collections(show_hidden=False):
    client = get_client()
    payload = {"show_hidden": str(show_hidden).lower()}
    result = client.get(f"{SERVER_URL}/collections", params=payload)
    return result.json()


def list_projects(collection_name: str, show_hidden: bool = False):
    client = get_client()
    payload = {"show_hidden": str(show_hidden).lower()}
    resp = client.get(f"{SERVER_URL}/projects/{collection_name}", params=payload)
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def create_project(
    collection_name: str,
    project_name: str,
    force: bool = False,
    storage_location: str = None,
):
    client = get_client()
    args = {"force": str(force).lower()}
    if storage_location:
        args["storage_location"] = storage_location

    result = client.post(
        f"{SERVER_URL}/create/{collection_name}/{project_name}", params=args
    )
    if result.status_code == 201:
        return result.json()
    elif result.status_code == 409:
        raise AlreadyExists(f"{result.json()}")
    else:
        raise GodataClientError(f"{result.status_code}: {result.text}")


def delete_project(collection_name: str, project_name: str, force: bool = False):
    client = get_client()
    payload = {"force": str(force).lower()}
    resp = client.delete(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}", params=payload
    )
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    elif resp.status_code == 403:
        raise Forbidden(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def load_project(collection_name: str, project_name: str):
    """
    Load a project into the server memory if it is not already loaded.
    """
    client = get_client()
    resp = client.post(f"{SERVER_URL}/load/{collection_name}/{project_name}")
    if resp.status_code == 200:
        print(resp.json())
        return True
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def drop_project(collection_name: str, project_name: str):
    """
    Signals to the server this client is done with this project. This may or may not
    actually drop the project from memory, depending on if other clients are using it.
    """
    client = get_client()
    try:
        resp = client.post(f"{SERVER_URL}/drop/{collection_name}/{project_name}")
    except requests.exceptions.ConnectionError:
        # The server is probably down, so this operation doesn't really matter
        return {}
    if resp.status_code == 200:
        return resp.json()
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def path_exists(collection_name: str, project_name: str, project_path: str):
    client = get_client()
    params = {"project_path": project_path}
    resp = client.get(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/exists", params=params
    )
    if resp.status_code == 200:
        return resp.json()
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def link_file(
    collection_name: str,
    project_name: str,
    project_path: str,
    file_path: str,
    metadata: dict = {},
    force: bool = False,
):
    client = get_client()
    params = {
        "project_path": project_path,
        "real_path": file_path,
        "force": str(force).lower(),
    }
    if set(metadata.keys()).intersection(set(params.keys())):
        raise GodataClientError(
            f"Metadata keys {set(metadata.keys())} conflict with parameter keys "
            f"{set(params.keys())}."
        )
    for k, v in metadata.items():
        try:
            params.update({str(k): str(v)})
        except TypeError:
            raise GodataClientError(
                "Metadata keys and values must be convertible strings."
            ) from None

    resp = client.post(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/files", params=params
    )
    if resp.status_code == 201:
        return resp.json()
    elif resp.status_code == 409:
        raise AlreadyExists(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def link_folder(
    collection_name: str,
    project_name: str,
    project_path: str,
    folder_path: str,
    recursive: bool = False,
):
    client = get_client()
    params = {
        "project_path": project_path,
        "real_path": folder_path,
        "type": "folder",
        "recursive": str(recursive).lower(),
    }
    resp = client.post(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/files", params=params
    )
    if resp.status_code == 201:
        return resp.json()
    elif resp.status_code == 409:
        raise AlreadyExists(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def list_project_contents(
    collection_name: str,
    project_name: str,
    project_path=None,
    show_hidden: bool = False,
):
    client = get_client()
    params = {"show_hidden": str(show_hidden).lower()}
    if project_path:
        params["project_path"] = project_path

    resp = client.get(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/list", params=params
    )
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def get_file(collection_name: str, project_name: str, project_path: str):
    client = get_client()
    params = {"project_path": project_path}
    resp = client.get(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/files", params=params
    )
    if resp.status_code == 200:
        return resp.json()
    else:
        raise NotFound(f"{resp.json()}")


def generate_path(collection_name: str, project_name: str, project_path: str):
    client = get_client()
    params = {"project_path": project_path}
    resp = client.get(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/generate",
        params=params,
    )
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def remove_file(collection_name: str, project_name: str, project_path: str):
    client = get_client()
    params = {"project_path": project_path}
    resp = client.delete(
        f"{SERVER_URL}/projects/{collection_name}/{project_name}/files", params=params
    )
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def export_tree(collection_name: str, project_name: str, output_path: Path):
    client = get_client()
    params = {"output_path": str(output_path)}
    resp = client.get(
        f"{SERVER_URL}/export/{collection_name}/{project_name}", params=params
    )
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")


def import_tree(collection_name: str, project_name: str, input_path: Path):
    client = get_client()
    params = {"input_path": str(input_path)}
    resp = client.get(
        f"{SERVER_URL}/import/{collection_name}/{project_name}", params=params
    )
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code == 404:
        raise NotFound(f"{resp.json()}")
    else:
        raise GodataClientError(f"{resp.status_code}: {resp.text}")



import socket

import requests
import urllib3

"""
An adapter for performing HTTP requests over a Unix domain socket. This code borrows
heavily from the Docker python client's implementation of the same thing:

https://github.com/docker/docker-py/blob/main/docker/transport/unixconn.py

There is a requests_unixsocket package, but it has not been updated in several years
and there is a breaking issue, so I included my own version.

I thought about using async HTTP requests, but I don't think it's necessary for this
project. If requests is good enough for the Docker client, it's good enough for me.
"""


class UnixHTTPConnection(urllib3.connection.HTTPConnection):
    def __init__(self, base_url, unix_socket, timeout=60):
        super().__init__("localhost", timeout=timeout)
        self.base_url = base_url
        self.unix_socket = unix_socket
        self.timeout = timeout

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect(self.unix_socket)
        self.sock = sock


class UnixHTTPConnectionPool(urllib3.connectionpool.HTTPConnectionPool):
    def __init__(self, base_url, socket_path, timeout=60, maxsize=10):
        super().__init__("localhost", timeout=timeout, maxsize=maxsize)
        self.base_url = base_url
        self.socket_path = socket_path
        self.timeout = timeout

    def _new_conn(self):
        return UnixHTTPConnection(self.base_url, self.socket_path, self.timeout)


class UnixHTTPAdapter(requests.adapters.HTTPAdapter):
    __attrs__ = requests.adapters.HTTPAdapter.__attrs__ + [
        "pools",
        "socket_path",
        "timeout",
        "max_pool_size",
    ]

    def __init__(self, socket_url, timeout=60, **kwargs):
        socket_path = socket_url.replace("http+unix://", "")
        if not socket_path.startswith("/"):
            socket_path = f"/{socket_path}"
        self.socket_path = socket_path
        self.timeout = timeout
        self.pools = urllib3._collections.RecentlyUsedContainer(
            dispose_func=lambda p: p.close()
        )
        super().__init__(**kwargs)

    def get_connection(self, url, proxies=None):
        with self.pools.lock:
            pool = self.pools.get(url)
            if pool:
                return pool

            pool = UnixHTTPConnectionPool(url, self.socket_path)
            self.pools[url] = pool

        return pool

    def request_url(self, request, proxies):
        # The select_proxy utility in requests errors out when the provided URL
        # doesn't have a hostname, like is the case when using a UNIX socket.
        # Since proxies are an irrelevant notion in the case of UNIX sockets
        # anyway, we simply return the path URL directly.
        # See also: https://github.com/docker/docker-py/issues/811
        return request.path_url



import click

from godata import server as srv


# top level godata command
@click.group()
def server():
    pass


# server subcommand
@server.command()
def start():
    """
    Start the godata server. This will start the server in the background and return
    """
    srv.start()


@server.command()
def stop():
    """
    Stop the godata server if it is running.
    """
    try:
        srv.stop()
    except Exception as e:
        print(e)


@server.command()
@click.option(
    "--upgrade",
    "-u",
    is_flag=True,
    help="Upgrade the server to the latest version.",
)
def install(upgrade):
    if upgrade:
        srv.upgrade()
    else:
        srv.install()


@server.command()
def uninstall():
    pass



from pathlib import Path

import pandas as pd


def get_pd_csv_writer(type_: pd.DataFrame):
    def pandas_csv_writer(df: pd.DataFrame, path: Path, **kwargs):
        if "index" not in kwargs:
            kwargs["index"] = False
        df.to_csv(path, **kwargs)

    pandas_csv_writer.__sufix__ = ".csv"
    return pandas_csv_writer


def get_pd_csv_reader(suffix=".csv") -> pd.DataFrame:
    def pandas_csv_reader(path: Path, **kwargs):
        return pd.read_csv(path, **kwargs)

    return pandas_csv_reader


def get_pd_parquet_writer(type_: pd.DataFrame):
    def pandas_parquet_writer(df: pd.DataFrame, path: Path, **kwargs):
        if "index" not in kwargs:
            kwargs["index"] = False
        df.to_parquet(path, **kwargs)

    pandas_parquet_writer.__sufix__ = ".parquet"
    return pandas_parquet_writer


def get_pd_parquet_reader(suffix=".parquet") -> pd.DataFrame:
    def pandas_parquet_reader(path: Path, **kwargs):
        return pd.read_parquet(path, **kwargs)

    return pandas_parquet_reader



import os


def handle_overwrite(link_result: dict):
    """
    This function handles the case when a file has been overwritten in the project
    and should be delted from disk. This happens whenver a file that was stored in
    the project is overwritten.
    """
    overwritten_file = link_result["overwritten"]
    if overwritten_file != "none":
        os.remove(overwritten_file)
        print(f"Removed {overwritten_file} from disk")



"""
Utility functions for importing and exporting projects.

We do this by creating a new temporary project, then recursively storing all
files in the project being exported. This creates an exact copy of the original project,
but with all files stored (and linked!) in the temporary project's storage. Then
we just zip it up and we're done.

The only complications is that files in the tree store their full path, and this path
will obviously change when the projected is imported. We need to either fix that or 
change the server to only store relative paths when the file is internal.

"""
import zipfile
from pathlib import Path

from godata.client.client import export_tree, import_tree
from godata.project import (
    GodataProject,
    create_project,
    delete_project,
    has_project,
    load_project,
)


def export_project(
    project_name: str,
    collection_name: str = "default",
    output_location=None,
    verbose=False,
) -> Path:
    if output_location and not output_location.is_dir():
        raise ValueError("Output location must be a directory")
    if not output_location:
        output_location = Path.cwd()

    source_project = load_project(project_name, collection_name)
    target_project = create_project(
        project_name, ".temp", storage_location=output_location
    )
    export_helper(source_project, target_project)
    # Now, zip up the temporary project
    zip_path = output_location / f"{project_name}.zip"
    expected_location = output_location / f".temp.{project_name}"
    if not expected_location.exists():
        raise RuntimeError("Something went wrong with the export")
    export_tree(collection_name, project_name, expected_location)
    with zipfile.ZipFile(zip_path, "w") as zip_file:
        # Recursively add all files in the temp project
        for f in expected_location.glob("**/*"):
            zip_file.write(f, f.relative_to(expected_location))
    # Clean up the temp project
    del target_project
    delete_project(project_name, ".temp", True)
    return zip_path


def export_helper(
    source_project: GodataProject,
    destination_project: GodataProject,
    project_path: str = None,
) -> None:
    folder_contents = source_project.list(project_path)
    files = folder_contents["files"]
    folders = folder_contents["folders"]
    if not project_path:
        project_path = ""
    for f in files:
        file_project_path = f"{project_path}/{f}"
        file_real_path = source_project.get(file_project_path, as_path=True)
        destination_project.store(file_real_path, file_project_path)
    for f in folders:
        export_helper(source_project, destination_project, f"{project_path}/{f}")


def import_project(
    zip_path: Path,
    project_name: str = None,
    collection_name: str = "default",
    output_location=None,
    verbose=False,
) -> None:
    if not zip_path.exists():
        raise ValueError("Zip file does not exist")
    if not project_name:
        project_name = zip_path.stem

    if has_project(project_name, collection_name):
        raise ValueError(
            f"Project {project_name} already exists in collection {collection_name}"
        )
    if not output_location:
        output_location = Path.cwd()
    if not output_location.is_dir():
        raise ValueError("Output location must be a directory")

    output_dir = output_location / f"{collection_name}.{project_name}"
    if output_dir.exists():
        raise ValueError(f"Output directory {output_dir} already exists")

    # unzip the project into the output location
    with zipfile.ZipFile(zip_path, "r") as zip_file:
        zip_file.extractall(output_dir)

    # Now, unzip the project
    # Now, import the project
    res = import_tree(collection_name, project_name, output_dir)
    if verbose:
        print(res)



# from . import SERVER_INSTALL_PATH
import os
import platform
import shutil
import subprocess
import zipfile

import requests

from godata import server

ENDPOINT = "https://sqm13wjyaf.execute-api.us-west-2.amazonaws.com/godata/download"
SERVER_INSTALL_PATH = "/usr/local/bin/godata_server"


def install(upgrade=False, version=None):
    """Install the godata server binary to /usr/local/bin/godata_server."""
    # detect the os this script is running on
    if upgrade and version is None:
        raise ValueError("Must specify the current version when upgrading.")

    params = {
        "install_type": "upgrade" if upgrade else "install",
        "os": platform.system().lower(),
    }

    if upgrade:
        params["current_version"] = version
    arch = platform.machine()
    if arch == "arm64":
        params["architecture"] = "aarch64"
    else:
        params["architecture"] = arch
    result = requests.get(ENDPOINT, params=params)
    if result.status_code == 200:
        response = result.json()
    else:
        print("Error: ", result.text)
        return

    url = response["url"]
    result = requests.get(url)
    # download the zip file
    # Stop the server if it is already running
    print("Stopping server if it is running...")
    server.stop()

    with open("godata_server.zip", "wb") as f:
        f.write(result.content)
    # unzip the file
    with zipfile.ZipFile("godata_server.zip", "r") as zip_ref:
        zip_ref.extractall()
    # move the binary to the install path
    shutil.move("godata_server", SERVER_INSTALL_PATH)
    # remove the zip file
    os.remove("godata_server.zip")
    # make the binary executable
    os.chmod(SERVER_INSTALL_PATH, 0o755)
    print("Restarting server...")
    server.start()
    if version is None:
        print(f"Successfully installed godata server version {response['version']}")
    else:
        print(
            f"Successfully upgraded godata server to version {response['version']} "
            f"from version {version}"
        )


def upgrade():
    """Upgrade the godata server binary to the latest version."""
    current_version = get_version().strip("\n")
    install(upgrade=True, version=current_version)


def get_version():
    try:
        return subprocess.check_output([f"{SERVER_INSTALL_PATH}", "--version"]).decode(
            "utf-8"
        )
    except FileNotFoundError:
        raise FileNotFoundError(
            "Unable to get godata server version: could not find the server binary. "
            "Please run `godata server install` first."
        )



import os
import signal
import subprocess
import time

import requests

from godata.client import client

from .install import SERVER_INSTALL_PATH, install, upgrade


def start():
    try:
        client.check_server()
    except client.GodataClientError as e:
        # Server is running, but client has a bad version
        print(e)
        return
    except requests.exceptions.ConnectionError:
        is_running = False
    else:
        is_running = True
    if is_running:
        print("Server is already running.")
        return

    try:
        subprocess.Popen([f"{SERVER_INSTALL_PATH}"], close_fds=True, shell=True)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Unable to start godata server: could not find the server binary. "
            "Please run `godata server install` first."
        )
    time.sleep(0.5)
    return True


def stop():
    try:
        server_pid = subprocess.check_output(["pgrep", "godata_server"])
    except subprocess.CalledProcessError:
        print("Server is not running.")
        return

    os.kill(int(server_pid), signal.SIGINT)


__all__ = ["start", "stop", "install", "upgrade"]



from pathlib import Path

import click

from godata import ie

from .utils import split_name


@click.command(name="import")
@click.argument("path", type=Path)
@click.argument("project_name", type=str)
@click.option("--storage", "-s", type=Path, default=None)
def import_project(path: Path, project_name: str, storage: Path):
    """
    Import a project from a directory.
    """
    name, collection = split_name(project_name)
    ie.import_project(path, name, collection, storage, verbose=True)


@click.command(name="export")
@click.argument("project_name", type=str)
@click.option("--output", "-o", type=Path, default=None)
def export_project(project_name: str, output: Path):
    """
    Export a project to a directory.
    """
    name, collection = split_name(project_name)
    output_path = ie.export_project(name, collection, output, verbose=True)
    click.echo(f"Project exported to {output_path}")



import inspect
from typing import Callable, ParamSpec, TypeVar

T = TypeVar("T")
P = ParamSpec("P")
# Signature variable


def sanitize_project_path(func: Callable[P, T]) -> Callable[P, T]:
    """
    A decorator to sanitize project paths. This will convert any argument
    named "project_path" to a string, and remove any leading or trailing slashes.

    This ensures anyhting that gets passed to the underlying project is of the
    correct form
    """

    def wrapper(*args, **kwargs) -> T:
        # Get the signature of the function

        if len(args) == 0 and len(kwargs) == 0:
            return func()
        sig = inspect.signature(func)
        # Figure out where the project_path argument is
        arg_index = None
        for i, name in enumerate(sig.parameters.keys()):
            if name == "project_path":
                arg_index = i

        if not arg_index:
            raise ValueError(
                f"No argument named 'project_path' found in function {func.__name__}"
            )
        project_path = None
        if len(args) > arg_index:
            # If the argument is passed as a positional argument
            project_path = args[arg_index]

        elif "project_path" in kwargs:
            # If the argument is passed as a keyword argument
            project_path = kwargs["project_path"]

        if project_path is not None and isinstance(project_path, str):
            if project_path.startswith("/"):
                # remove the leading slash
                project_path = project_path[1:]
            if project_path.endswith("/"):
                # remove the trailing alsh
                project_path = project_path[:-1]

        if arg_index is not None and arg_index < len(args):
            # If the argument is passed as a positional argument
            args = list(args)
            args[arg_index] = project_path
        else:
            # If the argument is passed as a keyword argument
            kwargs["project_path"] = project_path
        return func(*args, **kwargs)

    return wrapper



import numpy as np


def get_np_writer(type_: np.ndarray):
    def write_numpy_array(array: np.ndarray, path: str, **kwargs):
        np.save(path, array, **kwargs)

    write_numpy_array.__sufix__ = ".npy"
    return write_numpy_array


def get_np_reader(suffix=".npy") -> np.ndarray:
    def load_numpy_array(path: str, **kwargs):
        return np.load(path, **kwargs)

    return load_numpy_array



from pathlib import Path

import polars as pl


def get_pd_writer(type_: pl.DataFrame):
    def write_polars_csv(df: pl.DataFrame, path: Path, **kwargs):
        df.write_csv(path, **kwargs)

    write_polars_csv.__sufix__ = ".csv"
    return write_polars_csv


def get_pd_reader(suffix=".csv") -> pl.DataFrame:
    def read_polars_csv(path: Path, **kwargs):
        return pl.read_csv(path, **kwargs)

    return read_polars_csv



from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any

import portalocker
from loguru import logger

from godata.client import client
from godata.files import utils as file_utils
from godata.io import find_writer, get_typekey, godataIoException, try_to_read
from godata.utils import sanitize_project_path

__all__ = ["load_project", "list_projects", "create_project", "GodataProjectError"]


class GodataProjectError(Exception):
    pass


class GodataProject:
    """
    This is the main user-facing class for interacting with projects. It mostly a
    thin wrapper around an http client, but it also delegates file reading and writing
    tasks where appropriate.

    This class also provides docstrings for the underlying methods, such that all
    user-facing documentation can be done with sphinx.
    """

    def __init__(self, collection, name) -> GodataProject:
        self.collection = collection
        self.name = name

    def __del__(self):
        client.drop_project(self.collection, self.name)

    @sanitize_project_path
    def remove(self, project_path: str) -> bool:
        """
        Remove an file/folder at the given path. If a folder contains other
        files/folders, this will throw an error unless rucursive is set to True.
        """
        client.remove_file(self.collection, self.name, project_path)
        # will raise an error if it cannot be removed
        return True

    def get(
        self,
        project_path: str,
        as_path: bool = False,
        load_type: type = None,
        reader_kwargs: dict = {},
    ) -> Any:
        """
        Get an object at a given project path. This method will return a python object
        whenever possible. If godata doesn't know how to read in a file of this type,
        it will return a path. The path can also be returned explicitly by passing
        as_path = True.
        """
        file_info = self.get_metadata(project_path)
        path_str = file_info["real_path"]
        path = Path(path_str)
        if as_path:
            return path
        try:
            if load_type is not None:
                format = get_typekey(load_type)
            else:
                format = file_info.get("obj_type")

            with portalocker.Lock(str(path), "rb"):
                data = try_to_read(path, format, reader_kwargs)
            return data
        except godataIoException as e:
            logger.info(
                f"Could not find a reader for file {path}. Returning path instead."
                f"Error: {e}"
            )
            return path

    @sanitize_project_path
    def get_metadata(self, project_path: str) -> dict:
        """
        Get the metadata for a given file. This will return a dictionary of metadata
        for the file. If the file does not exist, this will throw an error.
        """
        file_info = client.get_file(self.collection, self.name, project_path)
        return file_info

    @sanitize_project_path
    def store(
        self,
        object: Any,
        project_path: str,
        overwrite=False,
        format: str = None,
        writer_kwargs: dict = {},
    ) -> bool:
        """
        Stores a given python object in godata's internal storage at the given path.
        Not having a writer defined in godata's python io module is not necessarily
        a failure case. Some objects can be converted easily into rust objects (or)
        actually ARE rust objects under the hood, and will be handled by the rust
        library. If a writer is not found by either python pyestor rust, this will throw
        an error.

        However one thing to note is that if a writer is found in python, it will
        always be used over a rust writer.
        """
        # First, see if the object is a path
        try:
            to_read = Path(object)
        except TypeError:
            to_read = object

        class_name = type(object).__name__
        module_name = type(object).__module__
        type_key = f"{module_name}.{class_name}"
        metadata = {"obj_type": type_key}

        # We link first, because it's better to have be tracking a file that doesn't
        # exist than to have a file that exists but isn't tracked.

        if isinstance(to_read, Path):
            try:
                obj = try_to_read(to_read)  # This can be very slow... Could be improved
                writer_fn, suffix = find_writer(obj, to_read)

            except godataIoException as e:
                logger.warning(
                    f"Could not find a reader for file {to_read}. The file will still"
                    "be stored, but godata will only be able to return a path."
                    f"Error: {e}"
                )
                storage_path = client.generate_path(
                    self.collection, self.name, project_path
                )
                storage_path = Path(storage_path)
                storage_path = storage_path.with_suffix(to_read.suffix)
                storage_path.parent.mkdir(parents=True, exist_ok=True)

                self.link(
                    storage_path,
                    project_path,
                    overwrite=overwrite,
                    metadata=metadata,
                    _force=True,
                )
                shutil.copy(to_read, storage_path)
                return True
        else:
            obj = object
            writer_fn, suffix = find_writer(object, format)
            if writer_fn is None:
                raise godataIoException(
                    f"No writer found for object of type {type(object)}"
                )

        if suffix is None:
            raise godataIoException(
                f"No writer found for object of type {type(object)}"
            )

        storage_path = client.generate_path(self.collection, self.name, project_path)

        storage_path = Path(storage_path)
        storage_path = storage_path.with_suffix(suffix)
        storage_path.parent.mkdir(parents=True, exist_ok=True)
        self.link(
            storage_path,
            project_path,
            overwrite=overwrite,
            _force=True,
            metadata=metadata,
        )
        with portalocker.Lock(str(storage_path), "wb"):
            writer_fn(obj, storage_path, **writer_kwargs)

        return True

    @sanitize_project_path
    def link(
        self,
        file_path: str,
        project_path: str,
        metadata: dict = {},
        recursive: bool = False,
        overwrite=False,
        _force=False,
    ) -> bool:
        """
        Add a file to the project. This will not actually move any data, just create
        a reference to the file.
        """

        fpath = Path(file_path)
        if not fpath.exists() and not _force:
            raise FileNotFoundError(f"Nothing found at {file_path}")
        fpath = fpath.resolve()

        if fpath.is_dir():
            result = client.link_folder(
                self.collection, self.name, project_path, str(fpath), recursive
            )
        else:
            try:
                result = client.link_file(
                    self.collection,
                    self.name,
                    project_path,
                    str(fpath),
                    metadata=metadata,
                    force=overwrite,
                )
            except client.AlreadyExists:
                raise GodataProjectError(
                    f"File already exists at {project_path}. Use overwrite=True to "
                    "overwrite it."
                )
        print(result["message"])
        file_utils.handle_overwrite(result)
        return True

    @sanitize_project_path
    def ls(self, project_path: str = None) -> None:
        """
        A basic ls utility for looking at projects. If a path is given, this will
        perform the ls in the folder at the given path. Otherwise, it will perform
        it in the project root.

        Just prints
        """
        contents = self.list(project_path)
        files = contents["files"]
        folders = contents["folders"]
        if not files and not folders:
            if project_path is None:
                print(f"No files or folders found in project '{self.name}'")
            else:
                print("No files or folders found at path '{}'".format(project_path))
            return

        if not project_path:
            header_string = f"Project `{self.name}` root:"
        else:
            header_string = f"{self.name}/{project_path}:"
        print(header_string)
        print("-" * len(header_string))
        for folder in folders:
            print(f"  {folder}/")
        for file in files:
            print(f"  {file}")

    @sanitize_project_path
    def has_path(self, project_path: str) -> bool:
        """
        Check if a given path exists in the project.
        """
        if not project_path:
            return True
        return client.path_exists(self.collection, self.name, project_path)

    @sanitize_project_path
    def list(self, project_path: str = None) -> dict[str, str]:
        """
        A basic ls utility for looking at projects. If a path is given, this will
        perform the ls in the folder at the given path. Otherwise, it will perform
        it in the project root.
        """
        return client.list_project_contents(self.collection, self.name, project_path)


def has_project(name: str, collection: str = "default") -> bool:
    """
    Check if a project exists in the given collection. If no collection is given,
    this will check the default collection.
    """
    if not has_collection(collection):
        return False
    projects = list_projects(collection, True, False)
    return name in projects


def has_collection(name: str) -> bool:
    """
    Check if a collection exists.
    """
    try:
        collections = list_collections(True, False)
        n_projects = len(list_projects(name, True))
    except client.NotFound:
        return False
    return name in collections and n_projects > 0


def create_project(
    name: str, collection: str = None, storage_location: str = None
) -> GodataProject:
    """
    Create a new project in the given collection. If no collection is given, this
    will create a project in the default collection. If the collection does not
    exist, it will be created.pyth

    """

    # Note, the manager will throw an error if the project already exists
    if collection is None:
        collection = "default"

    if has_project(name, collection):
        raise GodataProjectError(
            f"Project {name} already exists in collection {collection}"
        )
    # If a custom storage location exsts, we need to make sure it's valid.
    if storage_location is not None:
        storage_path = Path(storage_location)
        if not storage_path.exists():
            raise FileNotFoundError(
                f"Storage location {storage_location} does not exist"
            )
        if not storage_path.is_dir():
            raise NotADirectoryError(
                f"Storage location {storage_location} is not a directory"
            )
        project_dir = storage_path / f"{collection}.{name}"
        if project_dir.exists():
            raise GodataProjectError(
                f"This project does not exist, but a file or directory with the "
                f"correct name already exists at {project_dir}. Please remove "
                "this file or directory and try again."
            )
        project_dir.mkdir(parents=True, exist_ok=True)
        storage_location = project_dir
    response = client.create_project(
        collection, name, force=True, storage_location=storage_location
    )
    print(response)
    return GodataProject(collection, name)


def delete_project(name, collection="default", force=False) -> bool:
    """
    Remove a project and all data stored in godata's internal storage. At present,
    this explicitly forces the user the suply True as an argument as a confirmation.
    In the future, we may implement an option to output the internal files somewhere.
    """
    client.delete_project(collection, name, force)
    return True


def load_project(name, collection="default") -> GodataProject:
    # this raises an error if the project doesn't exist
    _ = client.load_project(collection, name)
    return GodataProject(collection, name)


def list_projects(collection="default", show_hidden=False, display=False) -> list[str]:
    projects = client.list_projects(collection, show_hidden)
    if display:
        print(f"Projects in collection `{collection or 'default'}`:")
        for p in projects:
            print(f"  {p}")
    return projects


def list_collections(show_hidden=False, display=False) -> list[str]:
    collections = client.list_collections(show_hidden)
    if display:
        print("Collections:")
        for c in collections:
            print(f"  {c}")
    return collections



import importlib
import inspect
import pkgutil
from pathlib import Path

_known_writers = {}
_known_readers = {}
_io_match = {}

"""


In general, we do not want godata to depend on every single package for
which we may provide input/output functionality. To facilitate this, i/o options are
seperated into individual files. If the library in question is not installed on the
user's machine, we skip over it.

Functions to get a given reader or writer must have the name "read" or "write" in them,
respectively.

Writers should take in a single argument with a type annotation. This type annotation
should be the type of an object the writer knows how to write. The "get_writer" function
(or whatever it's named) should return a function that takes in two arguments: the
object to write and a path to write to, in that order.

Readers should take in a single argument of type string, with the default value being
the suffix of the file that will be delegated to this reader. The "get_reader" function
(or whatever it's named) should return a function that takes in a single argument: the
path to the file to read. This function should return the object that was read in.
"""


class godataIoException(Exception):
    pass


def get_typekey(cls):
    class_name = cls.__name__
    module_name = cls.__module__
    type_key = f"{module_name}.{class_name}"
    return type_key


for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    try:
        spec = loader.find_spec(module_name)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except ImportError:
        continue
    children = [getattr(module, child) for child in dir(module)]
    for child in children:
        if hasattr(child, "__call__") and not child.__name__.startswith("__"):
            signature = inspect.signature(child)
            if "write" in child.__name__:
                if len(signature.parameters) != 1:
                    raise ValueError("Writer functions must take exactly one argument")
                par = list(signature.parameters.values())[0]
                if par.annotation == inspect._empty:
                    raise ValueError("Writer functions must have type annotations")
                elif not isinstance(par.annotation, type):
                    raise ValueError("Writer function annotations must be types")
                f_ = child(par.annotation)
                key = get_typekey(par.annotation)
                if key in _known_writers:
                    _known_writers[key].append(f_)
                else:
                    _known_writers.update({key: [f_]})
            elif "read" in child.__name__:
                if len(signature.parameters) != 1:
                    raise ValueError("Reader functions must take exactly one argument")
                par = list(signature.parameters.values())[0]
                return_type = signature.return_annotation
                if return_type == inspect._empty:
                    raise ValueError(
                        "Reader functions must have a return type annotation"
                    )

                return_type = get_typekey(signature.return_annotation)
                if par.default == inspect._empty:
                    raise ValueError(
                        "Reader functions must a default value for their "
                        "argument, which specifies a file suffix."
                    )
                elif not isinstance(par.default, str):
                    raise ValueError("Reader function annotations must be types")

                if (key := par.default) in _known_readers:
                    _known_readers[key].append((child(), return_type))

                else:
                    _known_readers.update({key: [(child(), return_type)]})


def get_known_readers():
    return _known_readers


def get_known_writers():
    return _known_writers


def try_to_read(path: Path, obj_type: type = None, reader_kwargs: dict = {}):
    readers = get_known_readers()
    suffix = path.suffix
    if suffix not in readers:
        raise godataIoException(f"No reader found for file type {suffix}")
    if obj_type is None:
        reader_fn = readers[suffix][0][0]

    else:
        for reader, type_ in readers[suffix]:
            if type_ == obj_type:
                reader_fn = reader
                break
        else:
            raise godataIoException(
                f"No reader found for file type {suffix} and object type {obj_type}."
                "Perhaps you need to install a library?"
            )

    return reader_fn(path, **reader_kwargs)


def find_writer(obj, format: str = None):
    obj_key = get_typekey(type(obj))
    writers = get_known_writers()
    if obj_key not in writers:
        raise godataIoException(f"No writer found for object type {obj_key}")
    available_writers = writers[obj_key]
    if len(available_writers) == 1 or not format:
        writer = available_writers[0]
    elif format:
        for write_fn in available_writers:
            if write_fn.__sufix__ == format:
                writer = write_fn
                break

        else:
            raise godataIoException(
                f"No writer found for object type {obj_key} and format {format}"
            )
    suffix = writer.__sufix__
    return writer, suffix


__all__ = ["try_to_read", "find_writer"]



def split_name(name: str) -> tuple:
    split = name.split("/")
    if len(split) == 1:
        project_name = split[0]
        collection = "default"
    elif len(split) == 2:
        project_name = split[1]
        collection = split[0]
    else:
        raise ValueError("Invalid project name.")
    return project_name, collection



import click

from . import ie, project
from .server import server


@click.group()
def main():
    """Command line interface for GoData."""
    pass


main.add_command(project.create)
main.add_command(project.link)
main.add_command(project.ls)
main.add_command(project.list)
main.add_command(project.get)
main.add_command(ie.export_project)
main.add_command(ie.import_project)
main.add_command(server)


if __name__ == "__main__":
    main()
